
// Range representation base class.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Pool' : Modified from Stroustrup.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Pool {
  struct Link {Link* next ; } ;

  unsigned esize ;
  Link* head ;

  void grow() ;

  int nalloc ;  // debug
public :
  void init(unsigned n) ;

  void* alloc() ;
  void free(void* b) ;
} ;

inline void* Pool :: alloc() {
  if(head == 0) grow() ;
  Link* p = head ;
  head = p->next ;
  nalloc++ ;  // debug
  return p ;
}

inline void Pool :: free(void* b) {
  Link* p = (Link*) b ;
  p->next = head ;
  head = p ;
  nalloc-- ;  // debug
}


#define POOL_NEW void* operator new(size_t) { return pool.alloc() ; }

#define POOL_DELETE void operator delete(void* p) { pool.free(p) ; }

#define POOL_STATE static Pool pool ; friend void RangeInit() ;

#define MEMORY_POOL_DEFN(T) Pool T :: pool ;

#define MEMORY_POOL_DECS public : POOL_NEW POOL_DELETE protected : POOL_STATE

void RangeInit() ;


enum Format {DIST_PRIMITIVE, DIST_COLLAPSED,
             DIST_BLOCK, DIST_CYCLIC, DIST_STEP} ;

#define MAX_LEVELS 2


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Location': The local subscript.                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

//class Range ;

struct Location : public Coord {
  //Location() {}

  //Location(const Range& x, int glb) ;

  int subs [MAX_LEVELS] ;

  // Cosmetic support for ghost regions:

  Location operator<<(const int shift) {
    Location shifted(*this) ;
    *(shifted.subs) -= shift ;
    return shifted ;
  }

  Location operator>>(const int shift) {
    Location shifted(*this) ;
    *(shifted.subs) += shift ;
    return shifted ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RangeRep': Base class for representation of ranges.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StrideRep ;

class RangeRep {
public :
  // Constructors and destructors

  RangeRep(const int _extent,
           RangeRep* _kernel, const int _level, Dimension _dimension) : 
      extent(_extent), kernel(_kernel), level(_level), dimension(_dimension),
      refCount(0) {}

  virtual ~RangeRep() {}

  // Subranges.

  virtual RangeRep* subrng(const int _extent,
                           const int _base, const int _stride) = 0 ;

  // Inquiries.

  int size() const {return extent ; }

  Dimension dim() const {return dimension ; }

  int lev() const {return level ; }
  virtual RangeRep* ker() const {return kernel ; }

  virtual RangeRep* tem() {return this ; }
  virtual int bas()  const {return 0 ; }
  virtual int str()  const {return 1 ; }

  virtual Format format() const = 0 ;

  // Subscripting and iteration.

  virtual int ker_glb(const int, int*) const {
    AdlibError("RangeRep :: ker_glb: member undefined on this range") ;
    return 0 ;
  }

  virtual void location(Location* loc, const int glb) const = 0 ;

  // Move error messages for following three into level-0 ranges?

  virtual RangeRep* shell() const {
    AdlibError("RangeRep :: shell: member undefined on this range") ;
    return 0 ;
  }

  virtual int block(const int, int*, int*) const {
    AdlibError("RangeRep :: block: member undefined on this range") ;
    return 0 ;
  }

  virtual int block_step(int*) const {
    AdlibError("RangeRep :: block_step: member undefined on this range") ;
    return 0 ;
  }

  virtual int crd(const int) const {
    AdlibError("RangeRep :: crd: member undefined on this range") ;
    return 0 ;
  }

  virtual int local(int*) const {
    AdlibError("RangeRep :: local: member undefined on this range") ;
    return 0 ;
  }

  virtual RangeRep* subker(int* ker_bas, int* ker_str) const {
    *ker_bas = 0 ;
    *ker_str = 1 ;
    return kernel ;
  }

  // Strides

  virtual int disp(const int sub) const {
    return sub ;
  }

  virtual int step(const int sub_stp) const {
    return sub_stp ;
  }

  virtual int offset(const Location* i) const {
    return i->subs [0] ;
  }

  virtual int width() const = 0 ;

//  virtual int volume() const {
//    if(level > 1)
//      return width() * kernel->volume() ;
//    else
//      return width() ;
//  }

  virtual int volume() const {return width() ;}

  // Reference counting.

  void incr() {refCount++ ; }
  void decr() {if((--refCount) == 0) delete this ; }
  
  // Interim support for PCRC Fortran interface:

  virtual int glb(const int sub, const int) const {return sub ; }

  // debugging

  virtual void show() = 0 ;
protected :
  int extent ;
  RangeRep* kernel ;
  int level ;
  Dimension dimension ;

private :
  int refCount ;

  // Interim support for PCRC Fortran interface:

  int g_lb, g_stride ;

friend class Range ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StrideRep': Base class for representation of strides.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

//class StrideRep {
//public :
//  // Constructors and destructors
//
//  StrideRep(StrideRep* _kernel, const int _length) : 
//      kernel(_kernel), length(_length), refCount(0) {
//    if(kernel) kernel->incr() ;
//  }
//
//  virtual ~StrideRep() {
//    if(kernel) kernel->decr() ;
//  }
//
//  // Inquiries.
//
//  int len() const {return length ; }
//
//  StrideRep* ker() const {return kernel ; }
//
//  virtual int disp(const int sub) const {
//    return length * sub ;
//  }
//
//  virtual int disp_step(const int sub_stp) const {
//    return length * sub_stp ;
//  }
//
//  // Reference counting.
//
//  void incr() {refCount++ ; }
//  void decr() {if((--refCount) == 0) delete this ; }
//  
//protected :
//  StrideRep* kernel ;
//  int length ;
//
//private :
//  int refCount ;
//
//friend class Stride ;
//
//MEMORY_POOL_DECS
//} ;


// ---------------------------------------------------
//
// Things to do:
//
//   Somewhere, run-time check that `MAX_LEVELS' isn't exceeded.
//
//   `disp', `step', `width' should be undefined for level-0 ranges?
//
//   Things that return `RangeRep*' should return `const RangeRep*'?

